from flask import url_for, render_template, redirect, abort, flash, request,\
        Blueprint, current_app, jsonify
from flask.ext.login import login_required
from flask.ext.wtf import Form
from wtforms.fields import StringField, SubmitField, HiddenField, SelectField,\
        Label
from wtforms.validators import InputRequired, AnyOf, NumberRange
from sqlalchemy.orm.exc import NoResultFound, MultipleResultsFound

from ..models import db
from ..auth.permissions import admin_permission
from ..auth import PermissionType
from ..auth.models import Division, Permission, Entity


blueprint = Blueprint('divisions', __name__)


@blueprint.route('/')
@login_required
@admin_permission.require()
def list_divisions():
    """Show a page listing all divisions.

    Accesible only to administrators.
    """
    return render_template('divisions.html', divisions=Division.query.all())


class AddDivisionForm(Form):
    name = StringField('Division Name', validators=[InputRequired()])
    submit = SubmitField('Create Division')


@blueprint.route('/add/', methods=['GET', 'POST'])
@login_required
@admin_permission.require()
def add_division():
    """Present a form for adding a view and also process that form.

    Only accesible to adminstrators.
    """
    form = AddDivisionForm()
    if form.validate_on_submit():
        division = Division(form.name.data)
        db.session.add(division)
        db.session.commit()
        return redirect(url_for('.division_detail', division_id=division.id))
    return render_template('form.html', form=form)


class ChangeEntity(Form):
    form_id = HiddenField(default='entity')
    id_ = HiddenField()
    name = StringField()
    permission = HiddenField(validators=[AnyOf(PermissionType.values())])
    action = HiddenField(validators=[AnyOf(('add', 'delete'))])


transformer_choices = [
    ('', ''),
    ('kill_timestamp', 'Kill Timestamp'),
    ('pilot', 'Pilot'),
    ('corporation', 'Corporation'),
    ('alliance', 'Alliance'),
    ('system', 'Solar System'),
    ('constellation', 'Constellation'),
    ('region', 'Region'),
    ('ship_type', 'Ship'),
    ('payout', 'Payout'),
    ('status', 'Request Status'),
]


class ChangeTransformer(Form):
    form_id = HiddenField(default='transformer')
    attribute = SelectField('Attribute', choices=transformer_choices)
    transformer = SelectField('Transformer', choices=[])


def transformer_choices(attr):
    default_transformers = [
        ('none', 'None'),
    ]
    choices = default_transformers
    if attr in current_app.url_transformers:
        for transformer in current_app.url_transformers[attr]:
            choices.append((transformer, transformer))
    return choices


@blueprint.route('/<int:division_id>/', methods=['GET', 'POST'])
@login_required
@admin_permission.require()
def division_detail(division_id):
    """Generate a page showing the details of a division.

    Shows which groups and individuals have been granted permissions to each
    division.

    Only accesible to administrators.

    :param int division_id: The ID number of the division
    """
    division = Division.query.get_or_404(division_id)
    if request.method == 'POST':
        if request.form['form_id'] == 'entity':
            form = ChangeEntity()
            if form.validate():
                if form.id_.data != '':
                    entity = Entity.query.get(form.id_.data)
                    if entity is None:
                        flash("No entity with ID #{}.".format(form.id_.data), 'error')
                else:
                    try:
                        entity = Entity.query.filter_by(name=form.name.data).one()
                    except NoResultFound:
                        flash("No entities with the name '{}' found.".
                                format(form.name.data), category='error')
                    except MultipleResultsFound:
                        flash("Multiple entities eith the name '{}' found.".
                                format(form.name.data), category='error')
                permission = PermissionType.from_string(form.permission.data)
                if form.action.data == 'add':
                    db.session.add(Permission(division, permission, entity))
                    flash("'{}' is now a {}.".format(entity,
                            permission.description.lower()), "info")
                elif form.action.data == 'delete':
                    Permission.query.filter_by(division=division,
                            permission=permission, entity=entity).delete()
                    flash("'{}' is no longer a {}.".format(entity,
                            permission.description.lower()), "info")
            else:
                abort(400)
        elif request.form['form_id'] == 'transformer':
            form = ChangeTransformer()
            attr = form.attribute.data
            form.transformer.choices = transformer_choices(attr)
            # Check form and go from there
            if form.validate():
                name = form.transformer.data
                if name == 'none':
                    division.transformers[attr] = None
                else:
                    # Get the specific map of transformers for the attribute
                    attr_transformers = current_app.url_transformers[attr]
                    # Get the new transformer
                    division.transformers[attr] = attr_transformers[name]
                    # Explicitly add the TransformerRef to the session
                    db.session.add(division.division_transformers[attr])
            else:
                abort(400)
        db.session.commit()
    return render_template(
            'division_detail.html',
            division=division,
            entity_form=ChangeEntity(),
            transformer_form=ChangeTransformer(),
    )


@blueprint.route('/<int:division_id>/transformers/')
@blueprint.route('/<int:division_id>/transformers/<attribute>/')
@login_required
@admin_permission.require()
def list_transformers(division_id, attribute=None):
    division = Division.query.get_or_404(division_id)
    if attribute is None:
        attrs = current_app.url_transformers.keys()
    else:
        attrs = (attribute,)
    choices = {}
    for attr in attrs:
        raw_choices = transformer_choices(attr)
        current = division.transformers.get(attr, None)
        if current is not None:
            choices[attr] = \
                    [(c[0], c[1], c[1] == current.name) for c in raw_choices]
        else:
            choices[attr] = \
                    [(c[0], c[1], False) for c in raw_choices]
    return jsonify(choices)



from flask import render_template
from flask.views import View
from flask.ext.login import login_required, current_user
from .. import db
from ..models import Request
from ..auth.models import User, Group, Permission, Division, Pilot


class RequestListing(View):
    """Abstract class for lists of :py:class:`~evesrp.models.Request`\s.
    """

    #: The template to use for listing requests
    template = 'list_requests.html'

    decorators = [login_required]

    def requests(self, division_id=None):
        """Returns a list :py:class:`~.Request`\s belonging to
        the specified :py:class:`~.Division`, or all divisions if
        ``None``. Must be implemented by subclasses, as this is an abstract
        method.

        :param int division_id: ID number of a :py:class:`~.Division`, or
            ``None``.
        :returns: :py:class:`~.models.Request`\s
        :rtype: iterable
        """
        raise NotImplementedError()

    def dispatch_request(self, division_id=None, page=1):
        """Returns the response to requests.

        Part of the :py:class:`flask.views.View` interface.
        """
        pager = self.requests(division_id).paginate(page, per_page=20)
        return render_template(self.template,
                pager=pager)

    @property
    def _load_options(self):
        """Returns a sequence of
        :py:class:`~sqlalchemy.orm.strategy_options.Load` objects specifying
        which attributes to load (or really any load options necessary).
        """
        return (
                db.Load(Request).load_only('id', 'pilot_id', 'division_id',
                    'system', 'ship_type', 'status', 'timestamp',
                    'base_payout'),
                db.Load(Division).joinedload('name'),
                db.Load(Pilot).joinedload('name'),
        )


class PersonalRequests(RequestListing):
    """Shows a list of all personally submitted requests and divisions they
    have permissions in.

    It will show all requests the current user has submitted.
    """

    template = 'personal.html'

    def requests(self, division_id=None):
        requests = Request.query\
                .join(User)\
                .filter(User.id==current_user.id)\
                .options(*self._load_options)
        if division_id is not None:
            requests = requests.filter(Request.division_id==division_id)
        requests = requests.order_by(Request.timestamp.desc())
        return requests


class PermissionRequestListing(RequestListing):
    """Show all requests that the current user has permissions to access.

    This is used for the various permission-specific views.
    """

    def __init__(self, permissions, statuses):
        """Create a :py:class:`PermissionRequestListing` for the given
        permissions and statuses.

        :param tuple permissions: The permissions to filter by
        :param tuple statuses: A tuple of valid statuses for requests to be in
        """
        self.permissions = permissions
        self.statuses = statuses

    def dispatch_request(self, division_id=None, page=1):
        if not current_user.has_permission(self.permissions):
            abort(403)
        else:
            return super(PermissionRequestListing, self).dispatch_request(
                    division_id, page)

    def requests(self, division_id=None):
        user_perms = db.session.query(Permission.id.label('permission_id'),
                Permission.division_id.label('division_id'),
                Permission.permission.label('permission'))\
                .filter(Permission.entity==current_user)
        group_perms = db.session.query(Permission.id.label('permission_id'),
                Permission.division_id.label('division_id'),
                Permission.permission.label('permission'))\
                .join(Group)\
                .filter(Group.users.contains(current_user))
        perms = user_perms.union(group_perms)\
                .filter(Permission.permission.in_(self.permissions))
        if division_id is not None:
            perms = perms.filter(Permission.division_id==division_id)
        perms = perms.subquery()
        requests = Request.query\
                .join(perms, Request.division_id==perms.c.division_id)\
                .filter(Request.status.in_(self.statuses))\
                .order_by(Request.timestamp.desc())\
                .options(*self._load_options)
        return requests



from flask import Request


class AcceptRequest(Request):

    _json_mimetypes = ['application/json',]

    _html_mimetypes = ['text/html', 'application/xhtml+xml']

    @property
    def _known_mimetypes(self):
        return self._json_mimetypes + self._html_mimetypes

    @property
    def wants_json(self):
        return self.accept_mimetypes.best_match(self._known_mimetypes) in \
            self._json_mimetypes

    @property
    def wants_html(self):
        return self.accept_mimetypes.best_match(self._known_mimetypes) in \
            self._html_mimetypes



from flask import url_for
from flask.json import JSONEncoder
from .models import Request, Action, Modifier
from .auth.models import User, Group, Division


class IterableEncoder(JSONEncoder):
    def default(self, o):
        try:
            iterator = iter(o)
        except TypeError:
            pass
        else:
            return list(o)
        return super(IterableEncoder, self).default(o)


class NamedEncoder(JSONEncoder):
    def default(self, o):
        try:
            ret = {
                'name': o.name,
                'id': o.id,
            }
        except AttributeError:
            pass
        else:
            if isinstance(o, User):
                ret['href'] = url_for('api.user_detail', user_id=o.id)
            elif isinstance(o, Group):
                ret['href'] = url_for('api.group_detail', group_id=o.id)
            elif isinstance(o, Division):
                ret['href'] = url_for('api.division_detail', division_id=o.id)
            if 'href' in ret:
                return ret
        return super(NamedEncoder, self).default(o)


class GrabbagEncoder(JSONEncoder):
    def default(self, o):
        try:
            ret = {'id': o.id}
        except AttributeError:
            pass
        else:
            if isinstance(o, Request):
                ret['href'] = url_for('requests.get_request_details',
                        request_id=o.id)
                attrs = ('killmail_url', 'kill_timestamp', 'pilot', 'alliance',
                    'corporation', 'submitter', 'division', 'status',
                    'base_payout', 'payout', 'details', 'id')
                for attr in attrs:
                    if attr == 'pilot':
                        ret[attr] = str(o.pilot)
                    elif attr == 'status':
                        ret[attr] = o.status.value
                    else:
                        ret[attr] = getattr(o, attr)
                ret['submit_timestamp'] = o.timestamp
                return ret
            elif isinstance(o, Action):
                ret['note'] = o.note or ''
                ret['timestamp'] = o.timestamp
                ret['user'] = o.user
                ret['type'] = o.type_.value
                return ret
            elif isinstance(o, Modifier):
                ret['user'] = o.user
                ret['timestamp'] = o.timestamp
                ret['note'] = o.note or ''
                if o.voided:
                    ret['void'] = {
                            'user': o.voided_user,
                            'timestamp': o.voided_timestamp,
                    }
                else:
                    ret['void'] = False
                ret['value'] = o.value
                ret['value_str'] = str(o)
                return ret
        return super(GrabbagEncoder, self).default(o)


# Multiple inheritance FTW
class SRPEncoder(IterableEncoder, NamedEncoder, GrabbagEncoder):
    pass
